__package__ = 'cana'
__title__ = u'CANAlization: Control & Redundancy in Boolean Networks'
__description__ = u'This package implements a series of methods used to study control, canalization and redundancy in Boolean networks.'

__author__ = """\n""".join([
    'Rion Brattig Correia <rionbr@gmail.com>',
    'Alex Gates <ajgates42@gmail.com>',
    'Xuan Wang <xw47@indiana.edu>',
    'Thomas Parmer <tjparmer@indiana.edu>',
    'Etienne Nzabarushimana <enzabaru@indiana.edu>',
    'Luis M. Rocha <rocha@indiana.edu>'
])

__copyright__ = u'2021, Correia, R. B., Gates, A., Rocha, L. M.'

__version__ = '0.1.1'
__release__ = '0.1.1'
#
__all__ = ['boolean_network', 'boolean_node']
